////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// The Fairisle Null Port Controller/////////////////////////////////////////////////// 
////  File Name: pnc_rtl.v                      
//// Designer: Jianping Lu
//// Description: The design is based on the Fairisle Null Port Controller designed by Cambridge ////                   University.The implementation which is done here is only for research purpose.
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
`define ip_idle                         3'b000
`define rx_wait                         3'b001
`define rx_store1                       3'b010
`define rx_store2                       3'b011
`define rx_data                         3'b100
`define tx_addr                         3'b101
`define tx_first5_data                  3'b110
`define tx_data                         3'b111

`define op_idle                         1'b0
`define op_rx_data                      1'b1
 
module pnc (framestart, 
                clk,
                npc_rst_n,
                rx_rd_req,                              
                rx_ip_soc,
                rx_ip_data,
                ctr_sz,
                ctr_id,
                ip_mem_data,
                mem_ip_data,
                ip_fab_data,
                fab_ip_ack,
                ip_mem_addr_r,
                ip_mem_addr_c,
                ip_mem_rd_req,
                ip_mem_wr_en,
                fab_op_data,
                op_fifo_data,
                op_fifo_wr_en,
                op_fab_ack,
                op_fifo_soc
                );

input framestart;       // A synchrounous signal, and equal to 64 clock cycles
input clk;  // clock signal
input npc_rst_n; // reset signal
input rx_rd_req; // indicate the receive board has sth to send
input rx_ip_soc;        // SOC signal from receive board to indicate the start of cell if it asserts
input [7 : 0]  rx_ip_data;       // Input data from receive board to the input port controller;
input ctr_sz; // For debugging, when set, cell will be put in row0, column 0 to 63
input ctr_id; // To prevent the data cell being written into the memory before new VCIs, FRB and
                        // PRB are written into the memory.     
output[7:0] ip_mem_data;      // The output data from the input port controller to memory. 
input[7:0]  mem_ip_data;        // The input data from the memory to the ip
output [7:0] ip_fab_data;    // The data from input port controller to the fabric
input   fab_ip_ack;  //acknowledge signal from the fabric to the input port controller
output [8 :0] ip_mem_addr_r; // row address of the memory given by IP
output[8 :0] ip_mem_addr_c; // column address of the memory given by IP
output ip_mem_rd_req;  //read request signal
output ip_mem_wr_en; // write enable signal 
input [7 : 0] fab_op_data; //data from fabric to output port controller
output [7:0] op_fifo_data; // data from output port controller to the fifo
output op_fifo_wr_en; // fifo write enable signal
output op_fab_ack;  // acknowledgment signal from output port controller to the fabric
output op_fifo_soc;  // start of cell signal from  op to fifo

reg [7 : 0] ip_mem_data;
reg [7 : 0] ip_fab_data;
reg [8 : 0] ip_mem_addr_r, ip_mem_addr_c;
reg ip_mem_rd_req, ip_mem_wr_en;
reg [7:0] op_fifo_data;
reg op_fab_ack;
reg op_fifo_soc;
reg op_fifo_wr_en;
reg ip_empty;
reg [2 : 0] ip_state_i;
reg op_state_i;
reg [5 : 0] ip_cell_cnt, op_cell_cnt;


// Process: ip_main_state
// Description: If ip_empty is '1', when framestart asserts, ip detects the rx_soc signal, 
// if it asserts, read the data following the signal, and latches the first two bytes (VCIs) 
// which will become the row and column address of a cell. when ip_empty is '0', it tries to 
// ransmit a cell. At four clock cycle after it transmits data, if it cannot receive 
// the acknowledgment signal, it stops transmitting. 
 
always @(posedge clk or npc_rst_n) begin 
if (npc_rst_n == 0) 
        ip_state_i = `ip_idle;
else begin 
        case (ip_state_i) 
                `ip_idle: if (ip_empty == 1 && framestart ==1 && rx_rd_req==1 && ctr_id == 0) 
                                ip_state_i = `rx_wait;
                           else if (ip_empty == 0 && framestart == 1 && ctr_id == 0)
                                ip_state_i = `tx_addr;
                           else ip_state_i = `ip_idle;
                `rx_wait: if (rx_ip_soc == 1) 
                                ip_state_i = `rx_store1;
                `rx_store1: ip_state_i = `rx_store2;
                `rx_store2: ip_state_i = `rx_data;
                `rx_data: if (ip_cell_cnt == 1)
                              ip_state_i = `ip_idle;
                           else
                                ip_state_i = `rx_data;
                `tx_addr: ip_state_i = `tx_first5_data;
                `tx_first5_data: if (ip_cell_cnt == (54-5) && fab_ip_ack == 0)
                                        ip_state_i = `ip_idle;
                                    else if (ip_cell_cnt==(54-5) && fab_ip_ack == 1)
                                        ip_state_i = `tx_data;
                                    else
                                      ip_state_i = `tx_first5_data;
                `tx_data: if (ip_cell_cnt == 1)
                                        ip_state_i = `ip_idle;
                           else
                                        ip_state_i = `tx_data;
        endcase
 end
end


// Process: ip_cell_cnt, define the ip_cell_cnt;

always @(npc_rst_n or posedge clk) begin
        if (npc_rst_n == 0) 
                ip_cell_cnt = 0;
        else if (ip_state_i == `rx_store2)
                ip_cell_cnt = 54;
        else if (ip_state_i == `tx_addr)
                ip_cell_cnt = 54;
        else if (ip_cell_cnt == 1 || ip_cell_cnt == 0)
                ip_cell_cnt = 0;
        else ip_cell_cnt = ip_cell_cnt -1;
end

// Process: op_cell_cnt, define the op_cell_cnt;

always @(npc_rst_n or posedge clk) begin
        if (npc_rst_n == 0) 
                op_cell_cnt = 0;
        else if (ip_state_i == `tx_first5_data
                 || ip_state_i == `tx_data)
                op_cell_cnt = `op_cell_cnt + 1;
        else 
                op_cell_cnt = 0;
end

// Process: op_cell_cnt, define the op_cell_cnt;

always @(npc_rst_n or posedge clk) begin
        if (npc_rst_n == 0) 
                op_cell_cnt = 0;
        else if (ip_state_i == `tx_first5_data
                 || ip_state_i == tx_data)
                op_cell_cnt = op_cell_cnt + 1;
        else 
                op_cell_cnt = 0;
end

// Process: define ip_mem_addr_c, ip_mem_addr_r

always @(npc_rst_n or posedge clk) begin 
        if (npc_rst_n == 0)  
        begin
                ip_mem_addr_c = 0;
                ip_mem_addr_r = 0;
        end
        else if (ctr_sz == 1)
        begin
                if (ip_state_i == `rx_store1
                        || ip_state_i == `rx_store2)
                begin
                        ip_mem_addr_r =  9'b000000000;
                        ip_mem_addr_c =  9'b000000100;
                end
                else if (ip_state_i == `rx_data 
                        || ip_state_i == `tx_first5_data
                        || ip_state_i == `tx_data)
                begin
                        ip_mem_addr_r = 9'b000000000;
                        ip_mem_addr_c[8:6] = 3'b000;
                        ip_mem_addr_c[5:0] = ip_mem_addr_c[5:0] + 1;
                end
                else if (ip_state_i == `tx_addr || ip_state_i == `ip_idle)
                begin
                        ip_mem_addr_r =  9'b000000000;
                        ip_mem_addr_c =  9'b000000000;
                end 
        end
        else    
        begin
                if (ip_state_i == `rx_store1)
                        ip_mem_addr_r[8:1] =  rx_ip_data;
                else if (ip_state_i == `rx_store2)
                begin
                        ip_mem_addr_r[0] = rx_ip_data[7];
                        ip_mem_addr_r[8:1] = ip_mem_addr_r[8:1];
                        ip_mem_addr_c[8:6] = rx_ip_data[6:4];
                        ip_mem_addr_c[5:0] = 6'b000100;
                end
                else if (ip_state_i == `rx_data)
                begin
                        ip_mem_addr_r  = ip_mem_addr_r;
                        ip_mem_addr_c[5:0] = ip_mem_addr_c[5:0] + 1;
                        ip_mem_addr_c[8:6] = ip_mem_addr_c[8:6];
                end
                else if (ip_state_i == `ip_idle) 
                begin   
                        ip_mem_addr_r = ip_mem_addr_r;
                        ip_mem_addr_c[8:6] = ip_mem_addr_c[8:6];
                        ip_mem_addr_c[5:0] = 6'b000000;
                end
                else if (ip_state_i == `tx_addr)
                begin
                        ip_mem_addr_r = ip_mem_addr_r;
                        ip_mem_addr_c = ip_mem_addr_c;
                end
                else if (ip_state_i == `tx_first5_data || 
                        ip_state_i == `tx_data)
                begin
                        ip_mem_addr_r = ip_mem_addr_r;
                        ip_mem_addr_c[8:6] = ip_mem_addr_c[8:6];
                        ip_mem_addr_c[5:0] = ip_mem_addr_c[5:0] + 1;
                end     
        end
end

// process: ip_mem_data, ip_mem_wr_en, define the data from ip to mem
always @(posedge clk or npc_rst_n) begin
        if (npc_rst_n == 0)
        begin
                ip_mem_data = 0;
                ip_mem_wr_en = 0;
        end
        else if (ip_state_i == `rx_data)
        begin
                ip_mem_data = rx_ip_data;
                ip_mem_wr_en = 1;
        end
        else begin
                ip_mem_data = ip_mem_data;
                ip_mem_wr_en = 0;
        end
end
        
// process: ip_fab_data, ip_mem_rd_req, define these two signals
always @(posedge clk or npc_rst_n) begin
        if (npc_rst_n == 0)
        begin
                ip_fab_data = 0;
                ip_mem_rd_req = 0;      
        end
        else if (ip_state_i == `tx_first5_data 
                || ip_state_i == `tx_data)
        begin
                ip_fab_data = mem_ip_data;
                ip_mem_rd_req = 1;
        end
        else 
        begin
                ip_fab_data = 9'b00000000;
                ip_mem_rd_req = 0;
        end
end

// Process: define ip_empty signal
always @(posedge clk or npc_rst_n)
begin 
        if (npc_rst_n == 0)
                ip_empty = 1;
        else if ((ip_state_i == `rx_data && ip_cell_cnt == 1)
                || (ip_state_i == `tx_data && ip_cell_cnt == 1))
                ip_empty = ~ip_empty;
        else
                ip_empty = ip_empty;
end

// Process: define op_state_i
always @(posedge clk or npc_rst_n)
begin
        if (npc_rst_n == 0)
                op_state_i = `op_idle;
        else
          begin 
                case (op_state_i)
                `op_idle: 
                         if (op_cell_cnt == 5 && fab_op_data[0] == 1)
                                op_state_i = `op_rx_data;
                         else op_state_i = `op_idle;
                `op_rx_data: 
                        if (op_cell_cnt == 57)  //5 + 52 
                                op_state_i = `op_idle;
                        else op_state_i = `op_rx_data;
                endcase
           end
end

// Process: define op_fifo_data, op_fifo_soc, op_fifo_wr_en, op_fab_ack

always @(posedge clk or npc_rst_n) begin
        if (npc_rst_n == 0) begin
                op_fifo_data = 0;
                op_fifo_soc = 0;
                op_fifo_wr_en = 0;
                op_fab_ack = 0;
        end
        else if (op_cell_cnt == 5 && fab_op_data[0] == 1)
                op_fifo_soc = 1;
        else if (op_state_i == `op_rx_data)
        begin
                op_fifo_soc = 0;
                op_fifo_data = fab_op_data;
                op_fifo_wr_en = 1;
                op_fab_ack = 1;
        end
        else
        begin
                op_fifo_soc = 0;
                op_fifo_data = 0;
                op_fifo_wr_en = 0;
                op_fab_ack = 0;
        end
end

endmodule
